###
### "THE BEER-WARE LICENSE":
### Alberto Ramos wrote this file. As long as you retain this 
### notice you can do whatever you want with this stuff. If we meet some 
### day, and you think this stuff is worth it, you can buy me a beer in 
### return. <alberto.ramos@cern.ch>
###
### file:    Inference.jl
### created: Mon Nov 22 14:49:04 2021
###                               

# Dalla Brida / Luscher coefficients of
# OMF integrator
const r1omf4 =  0.08398315262876693
const r2omf4 =  0.25397851084105950
const r3omf4 =  0.68223653357190910
const r4omf4 = -0.03230286765269967
const r5omf4 =  0.5-r1omf4-r3omf4
const r6omf4 =  1.0-2.0*(r2omf4+r4omf4)

const r1omf2 =  0.1931833275037836
const r2omf2 =  0.5
const r3omf2 =  1 - 2*r1omf2

struct IntrScheme{N, T}
    r::NTuple{N, T}
    eps::T
    ns::Int64    
end


omf2(::Type{T}, eps, ns) where T = IntrScheme{3,T}((r1omf2,r2omf2,r3omf2), eps, ns)
omf4(::Type{T}, eps, ns) where T = IntrScheme{6,T}((r1omf4,r2omf4,r3omf4,r4omf4,r5omf4,r6omf4), eps, ns)
leapfrog(::Type{T}, eps, ns) where T = IntrScheme{2,T}((0.5,1.0), eps, ns)


import Base.show
function Base.show(io::IO, int::IntrScheme{N,T}) where {N,T}

    if N == 2
        println(io, "LEAPFROG integration scheme")
    elseif N == 3
        println(io, "OMF2 integration scheme")
    elseif N == 6
        println(io, "OMF4 integration scheme")
    end
    println(io, "  - eps: ", int.eps)
    println(io, "  - ns:  ", int.ns)
    
    return nothing
end

function logpost(p, model::Model, data::BIData{NPT,T}, pr::Prior) where {NPT,T}

    S = ( (data.y[1] - model(data.x[1], p))/data.dy[1] )^2/2 + log(data.dy[1])
    for i in 2:NPT
        S = S + ( (data.y[i] - model(data.x[i], p))/data.dy[i] )^2/2 + log(data.dy[i])
    end

    S = S + pr(p)

    return S
end

struct BIWorkSpace
    cf_tape
    frc
    mom
    p1
    
    function BIWorkSpace(model::Model, dt::BIData, pr::Prior)
        frc     = zeros(nparameters(model))
        mom     = zeros(nparameters(model))
        p1      = zeros(nparameters(model))
        cf_tape = ReverseDiff.compile(ReverseDiff.GradientTape(p->logpost(p, model, dt, pr), frc))
        return new(cf_tape, frc, mom, p1)
    end
end

force!(frc, p, cf_tape) where {NPT,T} =  ReverseDiff.gradient!(frc, cf_tape, p)

function MD!(mom, p, int::IntrScheme{NI, T}, ws::BIWorkSpace) where {NI, T <: AbstractFloat}

    force!(ws.frc, p, ws.cf_tape)
    mom .= mom .- (int.r[1]*int.eps) .* ws.frc
    for i in 1:int.ns
        k   = 2
        off = 1
        for j in 1:NI-1
            p .= p .+ (int.eps*int.r[k]) .* mom
            if k == NI
                off = -1
            end
            k += off
            
            if (i < int.ns) && (k == 1)
                force!(ws.frc, p, ws.cf_tape)
                mom .= mom .- (2*int.r[k]*int.eps) .* ws.frc
            else
                force!(ws.frc, p, ws.cf_tape)
                mom .= mom .- (int.r[k]*int.eps) .* ws.frc
            end
            if k == NI
                off = -1
            end
            k += off
        end
    end
    
    return nothing
end

hamiltonian(mom, p, model::Model, data::BIData{NPT,T}, pr::Prior) where {T <: AbstractFloat, NPT} = sum(mom.^2)/2 + logpost(p, model, data, pr)

function HMC!(p, int::IntrScheme{NI, T}, model::Model, data::BIData{NPT,T}, pr::Prior, ws::BIWorkSpace) where {NI, T <: AbstractFloat, NPT}

    ws.p1 .= p
    Random.randn!(ws.mom)
    hini = hamiltonian(ws.mom, p, model, data, pr)

    MD!(ws.mom, p, int, ws)

    dh   = hamiltonian(ws.mom, p, model, data, pr) - hini
    pacc = exp(-dh)

    acc = true
    if pacc < 1.0
        r = rand()
        if pacc < r
            p .= ws.p1
            acc = false
        end
    end
    
    return dh, acc
end

function relax!(p, ns, model::Model, data::BIData{NPT,T}, pr::Prior, ws::BIWorkSpace; log=false) where {NPT,T}

    function line_search(f, x; tol = 0.001)

        x1 = x
        f1 = f(x1)
        
        h  = 0.1
        x2 = x1
        f2 = f1 + h
        n = 0
        while f2 >= f1
            h  = h/2
            x2 = x1 + h
            f2 = f(x2)

            n = n + 1
            if (n > 100)
                println(x1, " ", f1,  " ", x2,  " ", f2)
                error("No bracketing found")
            end
        end

        n = 0
        x3 = x1
        f3 = f2 - h
        while f3 <= f2
            h = 2*h
            x3 = x2 + h
            f3 = f(x3)

            n = n + 1
            if (n > 100)
                return x3
            end
        end

        n = 0
        while (abs(x1-x3) > tol) && (abs(f1-f2) < tol)
            x4 = x1 + (x3-x2)
            f4 = f(x4)
            if f4 < f2
                x1 = x2
                f1 = f2
                x2 = f4
                f2 = f4
            else
                x3 = x4
                f3 = f4
            end

            n = n + 1
            if (n > 100)
                return x2
            end

        end
        
        return x2
    end
    
    for i in 1:ns
        force!(ws.frc, p, ws.cf_tape)
        eps = line_search(h -> logpost(p .- h.*ws.frc, model, data, pr), 0.0)
        p .= p .- eps .* ws.frc
    end

    return nothing
end
